import { Injectable, Logger } from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { EntityNotFoundError, Repository } from 'typeorm';
import { PaginateQuery, Paginated, paginate } from 'nestjs-paginate';
import { ZoneCreateDto } from './dto/zone-create.dto';
import { ZoneUpdateDto } from './dto/zone-update.dto';
import { Zone } from './entities/zone.entity';
import { GroupCreateDto } from './dto/group-create.dto';
import { GroupUpdateDto } from './dto/group-update.dto';
import { Group } from './entities/group.entity';
import { Category } from './entities/category.entity';
import { CategoryCreateDto } from './dto/category-create.dto';
import { CategoryUpdateDto } from './dto/category-update.dto';

@Injectable()
export class ZgcService {
  constructor(
    private readonly logger: Logger,
    @InjectRepository(Zone)
    private zonesRepository: Repository<Zone>,
    @InjectRepository(Group)
    private groupsRepository: Repository<Group>,
    @InjectRepository(Category)
    private categoriesRepository: Repository<Category>,
  ) {}

  // -------------------- Zone --------------------
  // --
  createZone(zoneCreateDto: ZoneCreateDto): Promise<Zone> {
    // Use Repository.create() will copy the values to destination object.
    const zone = this.zonesRepository.create(zoneCreateDto);
    return this.zonesRepository.save(zone);
  }

  findZones(query: PaginateQuery): Promise<Paginated<Zone>> {
    return paginate(query, this.zonesRepository, {
      // loadEagerRelations: true,
      // relations: { groups: true },
      relations: ['groups'],
      sortableColumns: ['id', 'name', 'displayOrder'],
      defaultSortBy: [
        ['displayOrder', 'ASC'],
        ['id', 'ASC'],
        ['groups.displayOrder', 'ASC'],
        ['groups.id', 'ASC'],
      ],
      select: [
        'id',
        'name',
        'displayOrder',
        'groups.id',
        'groups.name',
        'groups.displayOrder',
        // 'groups.categories.id',
        // 'groups.categories.name',
        // 'groups.categories.displayOrder',
      ],
      // searchableColumns: ['id', 'name'],
      // filterableColumns: {
      //   'id': [
      //     FilterOperator.EQ,
      //     FilterSuffix.NOT,
      //   ],
      //   'name': [
      //     FilterOperator.EQ,
      //     FilterOperator.ILIKE,
      //     FilterSuffix.NOT,
      //   ],
      // },
    });
  }

  findZone(id: number): Promise<Zone> {
    return this.zonesRepository.findOneOrFail({ where: { id } });
  }

  async updateZone(id: number, zoneUpdateDto: ZoneUpdateDto): Promise<void> {
    await this.isZoneExist(id);
    await this.zonesRepository.update({ id }, zoneUpdateDto);
  }

  async deleteZone(id: number): Promise<void> {
    await this.isZoneExist(id);
    await this.zonesRepository.delete(id);
  }

  // Helper function: Check if the entity exist.
  // If entity does not exsit, return the Promise.reject()
  private async isZoneExist(id: number): Promise<void> {
    const cnt = await this.zonesRepository.countBy({ id });
    if (cnt > 0) {
      return;
    } else {
      return Promise.reject(new EntityNotFoundError(Zone, { where: { id } }));
    }
  }

  // -------------------- Group --------------------
  // --
  createGroup(groupCreateDto: GroupCreateDto): Promise<Group> {
    // Use Repository.create() will copy the values to destination object.
    const group = this.groupsRepository.create(groupCreateDto);
    return this.groupsRepository.save(group);
  }

  findGroups(query: PaginateQuery): Promise<Paginated<Group>> {
    return paginate(query, this.groupsRepository, {
      // loadEagerRelations: true,
      // relations: { groups: true },
      relations: ['zone', 'categories'],
      sortableColumns: ['id', 'name', 'displayOrder'],
      select: [
        'id',
        'name',
        'displayOrder',
        'zone.id',
        'zone.name',
        'categories.id',
        'categories.name',
        'categories.displayOrder',
      ],
      // searchableColumns: ['id', 'name'],
      // filterableColumns: {
      //   'id': [
      //     FilterOperator.EQ,
      //     FilterSuffix.NOT,
      //   ],
      //   'name': [
      //     FilterOperator.EQ,
      //     FilterOperator.ILIKE,
      //     FilterSuffix.NOT,
      //   ],
      // },
    });
  }

  findGroup(id: number): Promise<Group> {
    return this.groupsRepository.findOneOrFail({ where: { id } });
  }

  async updateGroup(id: number, groupUpdateDto: GroupUpdateDto): Promise<void> {
    await this.isGroupExist(id);
    await this.groupsRepository.update({ id }, groupUpdateDto);
  }

  async deleteGroup(id: number): Promise<void> {
    await this.isGroupExist(id);
    await this.groupsRepository.delete(id);
  }

  // Helper function: Check if the entity exist.
  // If entity does not exsit, return the Promise.reject()
  private async isGroupExist(id: number): Promise<void> {
    const cnt = await this.groupsRepository.countBy({ id });
    if (cnt > 0) {
      return;
    } else {
      return Promise.reject(new EntityNotFoundError(Group, { where: { id } }));
    }
  }

  // -------------------- Category --------------------
  // --
  createCategory(cateogyCreateDto: CategoryCreateDto): Promise<Category> {
    // Use Repository.create() will copy the values to destination object.
    const cateogy = this.categoriesRepository.create(cateogyCreateDto);
    return this.categoriesRepository.save(cateogy);
  }

  findCategories(query: PaginateQuery): Promise<Paginated<Category>> {
    return paginate(query, this.categoriesRepository, {
      // loadEagerRelations: true,
      // relations: { groups: true },
      relations: ['group'],
      sortableColumns: ['id', 'name', 'displayOrder'],
      select: ['id', 'name', 'displayOrder', 'group.id', 'group.name'],
      searchableColumns: ['id', 'name'],
      // filterableColumns: {
      //   'id': [
      //     FilterOperator.EQ,
      //     FilterSuffix.NOT,
      //   ],
      //   'name': [
      //     FilterOperator.EQ,
      //     FilterOperator.ILIKE,
      //     FilterSuffix.NOT,
      //   ],
      // },
    });
  }

  findCategory(id: number): Promise<Category> {
    return this.categoriesRepository.findOneOrFail({ where: { id } });
  }

  async updateCategory(
    id: number,
    categoryUpdateDto: CategoryUpdateDto,
  ): Promise<void> {
    await this.isCategoryExist(id);
    await this.categoriesRepository.update({ id }, categoryUpdateDto);
  }

  async deleteCategory(id: number): Promise<void> {
    await this.isCategoryExist(id);
    await this.categoriesRepository.delete(id);
  }

  // Helper function: Check if the entity exist.
  // If entity does not exsit, return the Promise.reject()
  private async isCategoryExist(id: number): Promise<void> {
    const cnt = await this.categoriesRepository.countBy({ id });
    if (cnt > 0) {
      return;
    } else {
      return Promise.reject(
        new EntityNotFoundError(Category, { where: { id } }),
      );
    }
  }
}
